import os
from time import time

from .payloads import Payload
from .presence import Presence
from .response import Response
from .utils import remove_none


class Activity:

    def __init__(self, client: Presence, pid: int = os.getpid(),
                 state: str = None, details: str = None,
                 start: int = None, end: int = None,
                 large_image: str = None, large_text: str = None,
                 small_image: str = None, small_text: str = None,
                 party_id: str = None, party_size: list = None,
                 join: str = None, spectate: str = None,
                 match: str = None, instance: bool = True):

        self.payload = Payload.set_activity(pid, state, details, start, end, large_image, large_text,
                                            small_image, small_text, party_id, party_size, join, spectate,
                                            match, instance, _rn=False)

        self.p_properties = ('pid', 'state', 'details', 'start', 'end', 'large_image', 'large_text', 'small_image', 'small_text', 'party_id', 'party_size', 'join', 'spectate', 'match', 'instance')
        self.response = Response.from_dict(self.payload.data)

        if not isinstance(client, Presence):
            raise NotImplementedError
        self.client = client

    def __setattr__(self, name, value):
        p = getattr(self, 'p_properties', None)
        r = getattr(self, 'response', None)
        if p and name in p:
            r.set_prop(name,value)

            payload = remove_none(self.response.to_dict())

            self.client.update(_donotuse=payload)
        else:
            self.__dict__[name] = value


    def __getattr__(self, name):
        p = object.__getattribute__(self, 'p_properties')
        r = object.__getattribute__(self, 'response')
        if p and name in p:
            return r.get_prop(name)

        return getattr(self, name, None)
    
    def end_in(self, time_until_end: int):
        if time_until_end < 0 or not isinstance(time_until_end, int):
            raise PyPresenceException('Must be a positive integer of how many seconds the activity will end in.')
        self.end = int(time())+time_until_end

# This SHOULD work, but is untested. Currently unsure if how I've done
# this is "right"... When you do Activity.state = 'my state here', it
# should call Presence.update() with a working payload. Had to do a bit
# of a hack to make it easier and ended up easing the Response class because it



import asyncio
import inspect
import json
import os
import struct
import sys
import tempfile
from typing import Union

from .exceptions import *
from .response import Response
from .payloads import Payload


class BaseClient:

    def __init__(self, client_id: str, **kwargs):
        pipe = kwargs.get('pipe', 0)
        loop = kwargs.get('loop', None)
        handler = kwargs.get('handler', None)
        self.async = kwargs.get('isasync', False)

        client_id = str(client_id)
        if sys.platform == 'linux' or sys.platform == 'darwin':
            self.ipc_path = (
                (os.environ.get('XDG_RUNTIME_DIR') or tempfile.gettempdir())
                + '/discord-ipc-' + str(pipe))
            self.loop = asyncio.get_event_loop()
        elif sys.platform == 'win32':
            self.ipc_path = r'\\?\pipe\discord-ipc-' + str(pipe)
            self.loop = asyncio.ProactorEventLoop()

        if loop is not None:
            self.loop = loop

        self.sock_reader = None  # type: asyncio.StreamReader
        self.sock_writer = None  # type: asyncio.StreamWriter

        self.client_id = client_id

        if handler is not None:
            if not inspect.isfunction(handler):
                raise PyPresenceException('Error handler must be a function.')
            args = inspect.getfullargspec(handler).args
            if args[0] == 'self':
                args = args[1:]
            if len(args) != 2:
                raise PyPresenceException('Error handler should only accept two arguments.')

            if self.async:
                if not inspect.iscoroutinefunction(handler):
                    raise InvalidArgument('Coroutine', 'Subroutine', 'You are running async mode - '
                                                                     'your error handler should be awaitable.')
                err_handler = self._async_err_handle
            else:
                err_handler = self._err_handle

            loop.set_exception_handler(err_handler)
            self.handler = handler

        if getattr(self, "on_event", None):  # Tasty bad code ;^)
            self._events_on = True
        else:
            self._events_on = False

    def _err_handle(self, loop, context: dict):
        result = self.handler(context['exception'], context['future'])
        if inspect.iscoroutinefunction(self.handler):
            loop.run_until_complete(result)

    async def _async_err_handle(self, loop, context: dict):
        await self.handler(context['exception'], context['future'])

    async def read_output(self):
        try:
            data = await self.sock_reader.read(1024)
        except BrokenPipeError:
            raise InvalidID
        code, length = struct.unpack('<II', data[:8])
        payload = json.loads(data[8:].decode('utf-8'))
        if payload["evt"] == "ERROR":
            raise ServerError(payload["data"]["message"])
        return Response.from_dict(payload, code=code)

    def send_data(self, op: int, payload: Union[dict, Payload]):
        if isinstance(payload, Payload):
            payload = payload.data
        payload = json.dumps(payload)
        self.sock_writer.write(
            struct.pack(
                '<II',
                op,
                len(payload)) +
            payload.encode('utf-8'))

    async def handshake(self):
        if sys.platform == 'linux' or sys.platform == 'darwin':
            self.sock_reader, self.sock_writer = await asyncio.open_unix_connection(self.ipc_path, loop=self.loop)
        elif sys.platform == 'win32' or sys.platform == 'win64':
            self.sock_reader = asyncio.StreamReader(loop=self.loop)
            reader_protocol = asyncio.StreamReaderProtocol(
                self.sock_reader, loop=self.loop)
            try:
                self.sock_writer, _ = await self.loop.create_pipe_connection(lambda: reader_protocol, self.ipc_path)
            except FileNotFoundError:
                raise InvalidPipe
        self.send_data(0, {'v': 1, 'client_id': self.client_id})
        data = await self.sock_reader.read(1024)
        code, length = struct.unpack('<ii', data[:8])
        if self._events_on:
            self.sock_reader.feed_data = self.on_event



class Response:

    def __init__(self, properties, code=None):
        valid_props = (list, tuple)
        if not isinstance(properties, valid_props):
            raise ValueError("Argument 'properties' should be type 'list' or 'tuple', not '{}'".format(type(properties)))

        self.code = code if code is not None else None
        self.properties = list(properties)

    def __repr__(self):
        def rend(d, l, root=None):
            root = root or []

            for key, val in d.items():
                if isinstance(val, dict):
                    rend(val, l, root+[key])
                else:
                    l.append("{par}{key} = {val}".format(par=''.join([parent+'.' for parent in root]), key=key, val=repr(val)))

        l = []
        rend(self._dict, l)
        return "[pypresence.Response\n    {}\n]".format('\n    '.join(l))

    def __str__(self):
        return self.__repr__()

    def to_dict(self):
        fdict = {}
        for p in self.properties:
            prop = getattr(self, p)
            if isinstance(prop, Response):
                fdict[p] = prop.to_dict()
            else:
                fdict[p] = getattr(self, p)
        return fdict

    @classmethod
    def from_dict(cls, from_dict: dict, code=None):

        if not isinstance(from_dict, dict):
            raise ValueError("Expected type 'dict' got type '{}' ".format(type(from_dict)))

        for key, value in from_dict.items():
            if isinstance(value, dict):
                value = Response.from_dict(value)
            setattr(cls, key, value)

        cls._dict = from_dict

        return cls(list(from_dict.keys()), code)

    def __getattr__(self, attr: str):  # Add shorthand for the payload's data
        data = getattr(self, 'data', None)
        if data and attr in self.data:
            return self.data.attr
        return self.attr

    def set_prop(self, name, value):
        for n in self.properties:
            if n == name:
                setattr(self,name,value)
                break
            elif isinstance(getattr(self, n), Response):
                getattr(self, n).set_prop(name,value)


    def get_prop(self, name):
        for n in self.properties:
            if n == name:
                return getattr(self,name)
                break
            elif isinstance(getattr(self, n), Response):
                r = getattr(self, n).get_prop(name)
                if r != None:
                    return r
